from PyQt5 import QtWidgets, uic
from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import QHBoxLayout, QVBoxLayout, QDockWidget, QWidget, QPushButton
import frequency_slider_widget as fs
import pzplot
import pyqtgraph as pg
import numpy as np
import sys
import serial
import argparse

#Append the PySP module for signal processing functions.
sys.path.append('..')
from PySP import Filters as filt
from PySP.fixed_point import Fixed_Point as fp

class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, com_port='COM10'):
        super(MainWindow, self).__init__()

        self.com_port = com_port

        self.setWindowTitle("Filter GUI")
        #Set the main window placement and size
        #self.setGeometry(300, 300,500, 500)

        #Import the frequency slider widget
        self.freq_slider_widget = fs.Frequency_Slider_Widget()

        #Load a plotting widget for the frequency response
        #pg.setConfigOption('background', pg.mkColor('#363535'))
        self.x_axis = pg.AxisItem('bottom')
        self.y_axis = pg.AxisItem('left')
        self.x_axis.setGrid(100)
        self.y_axis.setGrid(100)
        self.freq_resp_plot_widget = pg.PlotWidget(axisItems={'bottom':self.x_axis, 'left':self.y_axis})
        self.freq_resp_plot_widget.setBackground("#363535")
        self.freq_resp_plot_widget.setYRange(-15, 15)
        self.freq_resp_plot_widget.setXRange(np.log10(20), np.log10(44100/2.0))
        self.freq_resp_plot_widget.setLabel(axis="bottom", text="log10(Freq.) (Hz)")
        self.freq_resp_plot_widget.setLabel(axis="left", text="Gain (dB)")
        self.freq_resp_plot_widget.setTitle("Frequency Response")

        #Load a the pole zeros mapping widget
        self.pzmap = pzplot.PzPlot()

        #Button to send the coefficients via serial communication
        self.send_button = QPushButton("Send Coefficients")
        self.send_button.clicked.connect(self._send_coefficients) #Callback for button pressed

        #Create a layout to put both of the plot widgets horizontally
        self.plots_layout = QHBoxLayout()
        self.plots_layout.addWidget(self.freq_resp_plot_widget)
        self.plots_layout.addWidget(self.pzmap)

        #Make a main widget that houses all of the sub-widgets
        self.main_widget = QWidget()
        layout = QVBoxLayout()
        layout.addLayout(self.plots_layout)
        layout.addWidget(self.freq_slider_widget)
        layout.addWidget(self.send_button)

        self.main_widget.setLayout(layout)
        print(self.main_widget.size())
        self.setCentralWidget(self.main_widget)

        print(self.freq_slider_widget.minimumSizeHint())


        #Create a QTimer for updating the frequency response
        self.freq_resp_timer = QTimer()
        self.freq_resp_timer.timeout.connect(self._update_freq_resp)
        self.freq_resp_timer.start(100) #every 100 milliseconds

        self.f_s = 135000.0#Sampling frequency at 44100 h
        self.a = np.array([1, 0, 0]) #Biquad filter denominator coefficients
        self.b = np.array([1, 0, 0]) #Biquad filter numerator coefficients

        #Initialize serial interface
        if(type(self.com_port) != type(None)):
            self.ser = serial.Serial(self.com_port)

    def _update_freq_resp(self):
        """
        Callback function to update the frequency response (and plot).
        This callback function is called whenver the freq_resp_timer times out.

        Parameters:
            N/A
        Returns:
            N/A
        """
        #Get the slider values from the slider widget.
        G_dB = self.freq_slider_widget.gain_dB
        f_c = self.freq_slider_widget.center_frequency
        bandwidth = self.freq_slider_widget.bandwidth

        #Since the PySP function biquad_peakingEQ_filter takes the Q factor
        #instead of bandwidth, the bandwidth needs to be converted to Q factor.
        Q = f_c / bandwidth

        self.b, self.a = filt.biquad_peakingEQ_filter(G_dB, f_c, self.f_s, Q)

        w, mag_data, phase_data = filt.freq_plot_data(self.b, self.a, self.f_s)

        #Only display frequencies from 0 to f_s / 2
        w = (np.array_split(w, 2))[1]
        mag_data = (np.array_split(mag_data, 2))[1]
        phase_data = (np.array_split(phase_data, 2))[1]

        #convert from angular frequency to hZ. Also unnormalize the frequencies
        f = np.log10(w * self.f_s / (2*np.pi))
        self.freq_resp_plot_widget.plot(f, mag_data, clear=True, pen=pg.mkPen('#de7119'))

        #Get the poles and zeros from the biquad coefficients
        poles = self._quadratic_factor(self.a)
        zeros = self._quadratic_factor(self.b)

        self.pzmap.plot(poles=poles, zeros=zeros)
    def _quadratic_factor(self, coeffs):
        """
        Use the quadratic formula to find factors from a quadratic having form
        ax^2 + bx + c
        """
        a = complex(coeffs[0]); b = complex(coeffs[1]); c=complex(coeffs[2]);

        x_1 = (-1*b + np.sqrt(b**2. - 4.*a*c)) / (2.*a)
        x_2 = (-1*b - np.sqrt(b**2. - 4.*a*c)) / (2.*a)
        return np.array([x_1, x_2])
    def _send_coefficients(self):
        """
        Callback function from the send coefficients button to send the filter
        coefficients via UART to the fpga.

        The following is the packet format that the FPGA expects to recieve for
        loading filters.

        START_BYTE_1 -  8: <0XCA>
        START_BYTE_2 -  8: <0XAC>
        FILTER_NUM   -  8: <filter number>
        a_1          - 16: <a1_coeff_msb><a1_coeff_lsb>
        a_2          - 16: <a2_coeff_msb><a2_coeff_lsb>
        b_0          - 16: <b0_coeff_msb><b0_coeff_lsb>
        b_1          - 16: <b1_coeff_msb><b1_coeff_lsb>
        b_2          - 16: <b2_coeff_msb><b2_coeff_lsb>
        STOP_BYTE    -  8: <0XFA>
        BUFFER_BYTE  -  8: <0x00>
        BUFFER_BYTE  -  8: <0X00>

        Parameters:
            N/A
        Returns:
            N/A
        """

        #Convert the floating point coefficients to the desired fixed point format
        #Currently the fixed-point format is 16-bit total with 7 fractional bits.
        b = self.b / self.a[0] #Correctly normalize the coefficients
        a = self.a / self.a[0]

        #Convert the filter coefficients to fixed point
        b_fixed = [fp(i, W=16, F=14) for i in b]
        a_fixed = [fp(i, W=16, F=14) for i in a]

        print("a", a)
        print("b", b)
        packet = bytearray()
        packet.append(0xCA) #START_BYTE_1
        packet.append(0xAC) #START_BYTE_2
        packet.append(0x00) #FILTER_NUM: Currently only a single filter

        #Package up the 5 16-bIt coefficients
        for coeff in a_fixed[1:]:
            packet.append(int(coeff.bin_rep[0:8], base=2))
            packet.append(int(coeff.bin_rep[8:16], base=2))

        for coeff in b_fixed:
            packet.append(int(coeff.bin_rep[0:8], base=2))
            packet.append(int(coeff.bin_rep[8:16], base=2))

        packet.append(0xFA) #STOP BYTE
        packet.append(0x00) #BUFFER BYTES
        packet.append(0x00)

        if(type(self.com_port) != type(None)):
            self.ser.write(packet) #Send the data packet

if __name__ == '__main__':

    #Get the COM port from CMD.
    parser = argparse.ArgumentParser()
    parser.add_argument("--com_port", type=str,
            help="Input the serial COM Port Name.",
            required=False)
    args = parser.parse_args()

    #Start the PyQt application
    app = QtWidgets.QApplication([])
    main = MainWindow(com_port = args.com_port)
    main.show()
    sys.exit(app.exec_())
