from PyQt5 import QtWidgets
import pyqtgraph as pg
import sys
import numpy as np

class PzPlot(pg.PlotWidget):

    def __init__(self):
        super(PzPlot, self).__init__()
        self.plot_item = self.getPlotItem()
        self._init_unit_circle()
        self._init_pole_zero_plot()
        self.setBackground("#363535") #sets the background color

        self.x_axis_bottom = self.plot_item.getAxis('bottom')
        self.x_axis_bottom.setGrid(100)
        self.y_axis_left = self.plot_item.getAxis('left')
        self.y_axis_left.setGrid(100)
        self.setLabel(axis="bottom", text="Real(z)")
        self.setLabel(axis="left", text="Imag(z)")
        self.setTitle("Pole-Zero Plot")

    def _init_unit_circle(self):
        """
        Plot the unit circle for pole stability reference.

        Parameters:
            N/A
        Returns:
            N/A
        """
        #Generate x-values
        x_1 = np.linspace(-1, 1, 500)
        x_2 = np.linspace(1, -1, 500)
        x = np.concatenate((x_1, x_2))
        y_1 = np.sqrt(1 - (x_1**2))
        y_2 = -1 * y_1
        y = np.concatenate((y_1, y_2))

        #Add a new PlotDataItem to display the unit cirlce
        self.unit_circle_1 = pg.PlotDataItem(x, y)
        self.plot_item.addItem(self.unit_circle_1)
        self.unit_circle_1.setPen(pg.mkPen({"color": "00a8cc", "width": 1}))

    def _init_pole_zero_plot(self):
        """
        Initialize the pole zero plot by creating the plot item
        """
        self.pole_plot = pg.PlotDataItem(pen=None, symbol="x", symbolPen=pg.mkPen("ffa41b"),
                                         symbolBrush=pg.mkBrush("ffa41b"),symbolSize=15
                                         )
        self.zero_plot = pg.PlotDataItem(pen=None, symbol="o", symbolPen=pg.mkPen("ed4559"),
                                         symbolBrush=pg.mkBrush(None), symbolSize=15
                                         )
        self.plot_item.addItem(self.pole_plot)
        self.plot_item.addItem(self.zero_plot)

    def plot(self, poles=None, zeros=None):
        """
        Plot the poles and zeros.

        Parameters:
            N/A
        Returns:
            N/A
        """
        #Extrac the real and imaginary parts of the poles and zeros.
        if(type(poles) != type(None)):
            poles_real_part = np.real(poles)
            poles_imag_part = np.imag(poles)
            self.pole_plot.setData(poles_real_part, poles_imag_part)
        if(type(zeros) != type(None)):
            zeros_real_part = np.real(zeros)
            zeros_imag_part = np.imag(zeros)
            self.zero_plot.setData(zeros_real_part, zeros_imag_part)
    def set_plot_range(self, x_min, x_max, y_min, y_max):
        """
        Set the ranges of the entire plot to be fixed
        """
        self.setXRange(x_min, x_max)
        self.setYRange(y_min, y_max)

if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    pzplot = PzPlot()
    pzplot.show()
    pzplot.set_plot_range(-1.1, 1.1, -1.1, 1.1)

    poles = np.array([-1/np.sqrt(2) + 1j/np.sqrt(2), -1/np.sqrt(2) - 1j/np.sqrt(2)])
    zeros = np.array([-1 + 0.5j, -1 - 0.5j, -0.75])
    pzplot.plot(poles=poles, zeros=zeros)
    sys.exit(app.exec_())
