import numpy as np
import matplotlib.pyplot as plt

def lowpass_shelf_1st_order_filter(B_pi, f_c, f_s):
    """
    Determines the coefficients in the IIR form of the first
    order Lowpass-shelf filter. The z-domain(digital) filter
    is derived from the s-domain(analog) filter by using the
    bilinear transform. The numerator and denominator coefficients
    are given for the transfer function form.

               b_0 + b_1 * z^(-1)
    H(z) =  -------------------------
               1 + a_1 * z^-1

    Parameters:
        B_pi: This is the gain of the upper shelf. Note the lower first_order_LP_shelf
                is always at 0dB (mag 1).
        f_c: The frequency of saddle point (mid way between upper shelf and low shelf)
        f_s: The sampling frequency
    Returns:
        num: The numerator coefficients [b_0, b_1]
        den: The denominator coefficients [1, a_1]
    """
    #Intermediate variables to make calculations simpler
    w_c = 2 * np.pi * f_c
    T = 1 / (f_s * 2);
    a = T * B_pi * w_c;
    b = T * w_c;

    #Determine the zeros and poses to get the form
    #        G * (1 - z_1 * z^(-1))
    #H(z) = -------------------------
    #            (1 - p_1 * z^(-1))

    z_1 = (1 - a) / (1 + a); #Zero
    p_1 = (1 - b) / (1 + b); #Pole
    G = (1 + a) / (1 + b); # Gain

    #The standard transfer numberator and denominator can thus
    #be obtained to get the IIR form

    #           b_0 + b_1 * z^(-1)
    #H(z) =  -------------------------
    #           1 + a_1 * z^-1

    #Numberator of transfer function coefficients
    b_0 = G;
    b_1 = -1 * G * z_1;
    num = [b_0, b_1];

    #Denominator of transfer function coefficients
    a_1 = -1 * p_1;
    den = [1, a_1];

    return num, den
def freq_plot_data(b, a, f_s, dw=1/500):
    """
    Return the x-axis and y-axis data needed to plot the frequency response of
    the provided filter with form.

               b_0 + b_1 * z^(-1) + ... + b_M * z^(-M)
    H(z) =  ------------------------------------------
               a_0  + a_1 * z^-1 + ... + a_M * z^(-N)


    Parameters:
        b: The numerator coefficients given as [b_0, b_1, ..., b_M]
        a: The denominator coefficients given as [a_0, a_1, ..., a_N]
        f_s: The sampling frequency
        dw: step size between omega points.
    Returns:
        w: The x-axis (domain). This from -pi to pi with step size dw.
        mag_data: The magnitude frequency response of the filter. (as numpy array)
        phase_data: The phase reponse of the filter. (as numpy array)
    """
    w = np.arange(-np.pi, np.pi, dw) #omega (domain)
    M = len(b) #Order of the numerator (number of zeros)
    N = len(a) #Order of the denominator (number of poles)
    num_exps = np.linspace(0, M - 1,  M)
    den_exps = np.linspace(0, N - 1, N)

    #Compute the numerator and denominator of the filter by substituting z with e^(jw)
    num = np.array([(b[index] * np.exp(1j * i * w)) for index, i  in enumerate(num_exps)])
    num = np.sum(num, axis=0)

    den = np.array([(a[index] * np.exp(1j * i * w)) for index, i  in enumerate(den_exps)])
    den = np.sum(den, axis=0)

    H =  num / den

    mag_data = 20 * np.log10(np.absolute(H))
    phase_data = np.angle(H)

    return w, mag_data, phase_data

def freq_plot(b, a, f_s, dw=1/500, fig=-1):
    """
    Plot the frequency response of a digital filter given form
               b_0 + b_1 * z^(-1) + ... + b_M * z^(-M)
    H(z) =  ------------------------------------------
               a_0  + a_1 * z^-1 + ... + a_M * z^(-N)

    Parameters:
        b: The numerator coefficients given as [b_0, b_1, ..., b_M]
        a: The denominator coefficients given as [a_0, a_1, ..., a_N]
        f_s: The sampling frequency
        dw: distance between omega points.
    Returns:
        fig: The matplot lib figure with the frequency plot
    """
    #Create a new plotting figure
    if(fig == -1):
        fig = plt.figure()
        fig.canvas.set_window_title("Frequency Response")
    mag_plt = fig.add_subplot(1, 1, 1)
    phase_plt = mag_plt.twinx()

    w = np.arange(1E-4, np.pi, dw)
    M = len(b) #Order of the numerator (number of zeros)
    N = len(a) #Order of the denominator (number of poles)
    num_exps = np.linspace(0, M - 1,  M)
    den_exps = np.linspace(0, N - 1, N)

    #Compute the numerator and denominator of the filter by substituting z with e^(jw)
    num = np.array([(b[index] * np.exp(1j * i * w)) for index, i  in enumerate(num_exps)])
    num = np.sum(num, axis=0)

    den = np.array([(a[index] * np.exp(1j * i * w)) for index, i  in enumerate(den_exps)])
    den = np.sum(den, axis=0)

    H =  num / den

    #Plot the magnitude of the frequency plot
    mag_plt.plot(np.log10(w * f_s / (2*np.pi)), 20*np.log10(np.absolute(H)), label="Mag.")
    mag_plt.set_xlabel("log10(Frequency) Hz")
    mag_plt.set_ylabel("20log10(|H(e^(-jw))) dB")

    #Plot the phase of the frequency plot
    #phase_plt.plot(np.log10(w * f_s / (2*np.pi)), np.angle(H), 'r--', label="Phase")
    #phase_plt.set_xlabel("log10(Frequency) Hz")
    #phase_plt.set_ylabel("Phase. rad")
    #fig.align_labels()
    fig.legend()

    return fig

def filter_df1(b, a, x):
    """
    Given the coefficients of a digital filter given form

               b_0 + b_1 * z^(-1) + ... + b_M * z^(-M)
    H(z) =  ------------------------------------------
               a_0  + a_1 * z^-1 + ... + a_N * z^(-N)

    This function uses the "Direct Form I" Structure of filtering. The filtering
    is using the time domain form

    y[n] = (b_0/a_0)*x[n] + (b_1/a_0)*x[n-1] + ... + (b_M/a_0)*x[n - M]
            - (a_1/a_0)*y[n-1] - ... - (a_N/a_0)*y[n-N]

    Parameters:
        b: The numerator coefficients given as [b_0, b_1, ..., b_M]
        a: The denominator coefficients given as [a_0, a_1, ..., a_N]
        x: Signal: A numpy array of shape (1,N) that will be filtered
    Returns:
        y: Signal: A numpy array of shape (1,N) that is the filtered x signal.
    """
    #Get the length of the input sequence
    x = np.reshape(np.array(x), (-1,))
    L = x.shape[0]

    #Normalize all of the coefficients with a_0
    a_0 = a[0]
    a = np.reshape(np.array(a) / a_0, (-1,))
    N = a.shape[0]
    a = a[1:]

    b = np.reshape(np.array(b) / a_0, (-1,))
    M = b.shape[0]

    #Initialize y as a sequence
    y = np.zeros(L)
    y_temp = np.zeros(N-1)
    x_temp = np.zeros(M)

    for n in range(L):
        x_temp[1:M] = x_temp[0:M-1]
        x_temp[0] = x[n]
        z = np.dot(-1 * a,y_temp)
        v = np.dot(b, x_temp)

        y[n] = v + z
        y_temp[1:N-1] = y_temp[0:N-2]
        y_temp[0] = y[n]

    return y

def biquad_peakingEQ_filter(G_dB, f_c, f_s, Q):
    """
    Calculate the biquad filter coefficients for a digital filter of form.

               b_0 + b_1 * z^(-1) + b_2 * z^(-2)
    H(z) =  ------------------------------------------
               a_0  + a_1 * z^-1 + a_2 * z^(-2)


    Parameters:
        G_dB: The peaking gain of the filter in dB.
        f_c: The center frequency (peaking frequency) in Hz
        f_s: The sampling frequency in Hz
        Q: The quality factor

    Returns:
        Two numpy arrays containing the numerator and denominator coefficients.
        These are returned in form [b_0, b_1, b_2], [a_0, a_1, a_2].
    """

    #Calculate intermediate variables
    A = 10.0**(G_dB / 40.0)

    #Normalized center frequency
    w_0 = (2. * np.pi * f_c) / f_s
    p = np.sin(w_0); q = np.cos(w_0)

    alpha = (p / (2.0*Q))

    b_0 = 1  + (alpha*A)
    b_1 = -2 * q
    b_2 = 1 - (alpha*A)
    a_0 = 1 + (alpha/A)
    a_1 = -2*q
    a_2 = 1 - (alpha/A)

    return np.array([b_0, b_1, b_2]), np.array([a_0, a_1, a_2])
