import numpy as np

def binary_str_to_decimal(x, W, F):
    """
    Convert a binary string to the 2's complement fixed point decimal.

    Parameter:
        x: A string of characters of which can only be '1's or '0's.
        W: The total width of the input binary string.
        F: The fractional width of the binary string. Number of fractional bits.

    Returns:
        The decimal representation of the 2's complement fixed point decimal.
    """
    accumulator = 0
    if(x[0] == '1'): # Negative number
        accumulator = -1 * 2**(W - F - 1)

    for i in range(1, W):
        if(x[i] == '1'):
            accumulator = accumulator + 2**(W - F - 1 - i)

    return accumulator



class Fixed_Point(object):

    def __init__(self, x=0, W=16, F=0):
        """
        Initialize the fixed point number given a floating point decimal
        """
        self._initialize(x, W, F)

    def _initialize(self, x, W, F):
        """
        """
        self.raw_x = x
        self.W = W
        self.F = F

        #Determine some parameters based on dimensions
        self.resolution = 2**(-1 * self.F)
        #Range of values possible
        self.scale_max = (2**(self.W - 1)) - 1
        self.scale_min = -1 * (2**(self.W - 1))
        self.x_max = self.scale_max * self.resolution
        self.x_min = self.scale_min * self.resolution
        if((self.raw_x > self.x_max) or (self.raw_x < self.x_min)):
            raise FixedPointIncorrectDim("Cannot Represent desired decimal with fixed point dimensions")

        #Convert the floating point decimal to fixed point
        self._cvt_to_fixed_point()

    def _cvt_to_fixed_point(self):
        """
        Convert the the floating point decimal to 2s complement fixed point.

        Parameters:
            N/A
        Returns:
            N/A
        """
        temp = self.raw_x * (2**self.F)

        self.scale = np.round(temp).astype(int)
        self.bin_rep = np.binary_repr(self.scale, width=self.W)
        self.bin_rep_dec_point = self.bin_rep[:(self.W-self.F)] + '.' + self.bin_rep[self.W - self.F:]

    def decimal(self):
        """
        Takes the binary representation and outputs the decimal representation.
        This is typically a rounded version of the original x

        Parameters:
            N/A
        Returns:
            The decimal representation of type float
        """
        return (self.scale * self.resolution)

    def resize(self, W, F):
        """
        Resize the fixed point representations to the new dimensions.

        Parameters:
            W: The total bit width of the fixed point representation
            F: The number of fractional bits
        """
        x = self.decimal()
        self._initialize(x, W, F)
        return self

    def __add__(self, b):
        """
        Addition of fixed point numbers.
        """
        #Make sure the two numbers being added have the same dimensions
        if(self.W != b.W):
            raise FixedPointIncorrectDim("Width's are not equivalent")

        if(self.F != b.F):
            raise FixedPointIncorrectDim("Fractional lengths are not equivalent")

        c = self.scale + b.scale

        if((c > self.scale_max) or (c < self.scale_min)):
            raise FixedPointOverFlow("Addition results in an overflow")

        c = c * (2**(-1 * self.F))

        return(Fixed_Point(c, self.W, self.F))

    def __sub__(self, b):
        """
        Subraction of fixed point numbers
        """
        #Make sure the two numbers being added have the same dimensions
        if(self.W != b.W):
            raise FixedPointIncorrectDim("Width's are not equivalent")

        if(self.F != b.F):
            raise FixedPointIncorrectDim("Fractional lengths are not equivalent")

        c = self.scale - b.scale

        if((c > self.scale_max) or (c < self.scale_min)):
            raise FixedPointOverFlow("Subtraction results in an overflow")

        c = c * (2**(-1 * self.F))

        return(Fixed_Point(c, self.W, self.F))

    def __mul__(self, b):
        """
        Multiply the fixed point numbers.
        Note Multiplication results in a Width 2 times the inputs widths
        and a number of fractional parts 2 times the number of input
        fractional parts.
        """
        #Make sure the two numbers being added have the same dimensions
        if(self.W != b.W):
            raise FixedPointIncorrectDim("Width's are not equivalent")

        if(self.F != b.F):
            raise FixedPointIncorrectDim("Fractional lengths are not equivalent")

        W = 2 * self.W
        F = 2 * self.F
        c = self.scale * b.scale
        c = c * (2**(-1 * F))

        return(Fixed_Point(c, W, F))


    def __str__(self):
        """
        The string that should be printed when a print statement is called
        """
        return self.bin_rep_dec_point


class Error(Exception):
    """Base Class for exceptions in the module"""
    pass

class FixedPointOverFlow(Error):
    """"Exception raised when an overflow occurs in the fixed point number arithmetic"""

class FixedPointIncorrectDim(Error):
    """Exception for incorrect dimensions of inputs for fixed point arithmetic"""

    def __init__(self, message):
        self.message = message


def add(a, b, allow_overflow=False):
    """
    This is a special add that can allow for overflow error. Useful for
    simulation purposes of digital systems.

    Parameters:
        b: The
    """
    if(not allow_overflow):
        return(a + b)

    #Make sure the two numbers being added have the same dimensions
    if(a.W != b.W):
        raise FixedPointIncorrectDim("Width's are not equivalent")

    if(a.F != b.F):
        raise FixedPointIncorrectDim("Fractional lengths are not equivalent")

    c = a.scale + b.scale
    c = c * (2**(-1 * a.F))
    c_fixed = Fixed_Point(c, a.W + 1, a.F)

    #Cut of the first bit to give effect of overflow error
    c_overflowed = binary_str_to_decimal(c_fixed.bin_rep[1:], a.W, a.F)
    return(Fixed_Point(c_overflowed, a.W, a.F))

def sub(a, b, allow_overflow=False):
    """
    """
    if(not allow_overflow):
        return(a - b)

    #Make sure the two numbers being added have the same dimensions
    if(a.W != b.W):
        raise FixedPointIncorrectDim("Width's are not equivalent")

    if(a.F != b.F):
        raise FixedPointIncorrectDim("Fractional lengths are not equivalent")

    c = a.scale - b.scale
    c = c * (2**(-1 * a.F))
    c_fixed = Fixed_Point(c, a.W + 1, a.F)

    #Cut of the first bit to give effect of overflow error
    c_overflowed = binary_str_to_decimal(c_fixed.bin_rep[1:], a.W, a.F)
    return(Fixed_Point(c_overflowed, a.W, a.F))

if __name__ == "__main__":
    num_1 = Fixed_Point(x=3, W=3, F=0)
    num_2 = Fixed_Point(x=3, W=3, F=0)

    num_3 = num_1.add(num_2, allow_overflow=True)
    print(num_3)
    print(num_3.decimal_represented())
